<?php

namespace Drupal\homebox\Controller;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\homebox\Entity\HomeboxInterface;

/**
 * The Homebox controller.
 *
 * @package homebox\Controller
 */
class HomeboxController extends ControllerBase {

  /**
   * Check access to Homebox entity.
   *
   * @param \Drupal\homebox\Entity\HomeboxInterface $homebox
   *   Homebox entity.
   * @param \Drupal\Core\Session\AccountInterface $user
   *   User entity.
   *
   * @return \Drupal\Core\Access\AccessResultAllowed|\Drupal\Core\Access\AccessResultForbidden
   *   Access result.
   */
  public function accessGeneral(HomeboxInterface $homebox, AccountInterface $user = NULL) {
    if ($this->currentUser()->hasPermission('bypass homeboxes access')) {
      return AccessResult::allowed();
    }
    // If the user wants to access another users homebox-page,
    // check if the user has the right permission:
    if ($user->id() != $this->currentUser()->id()) {
      if (!$this->currentUser()->hasPermission('bypass homeboxes access')) {
        return AccessResult::forbidden();
      }
    }
    $user_roles = $user->getRoles();
    $restrictedRoles = $homebox->getRestrictedRoles();
    if (!empty($restrictedRoles)) {
      if (empty(array_intersect($user_roles, $restrictedRoles))) {
        return AccessResult::forbidden();
      }
    }
    if ($user->hasPermission('use homebox')) {
      return AccessResult::allowed();
    }

    return AccessResult::forbidden();
  }

  /**
   * Check access to homebox dashboard page form.
   *
   * @param \Drupal\homebox\Entity\HomeboxInterface $homebox
   *   Homebox entity.
   * @param \Drupal\Core\Session\AccountInterface $user
   *   User entity.
   *
   * @return \Drupal\Core\Access\AccessResultAllowed|\Drupal\Core\Access\AccessResultForbidden
   *   Access result.
   */
  public function accessDashboardPage(HomeboxInterface $homebox, AccountInterface $user = NULL) {
    if (!$homebox->getDashboardPageEnabled()) {
      return AccessResult::forbidden();
    }
    return $this->accessGeneral($homebox, $user);
  }

  /**
   * Check access to homebox user tab form.
   *
   * @param \Drupal\homebox\Entity\HomeboxInterface $homebox
   *   Homebox entity.
   * @param \Drupal\Core\Session\AccountInterface $user
   *   User entity.
   *
   * @return \Drupal\Core\Access\AccessResultAllowed|\Drupal\Core\Access\AccessResultForbidden
   *   Access result.
   */
  public function accessDashboardUserTab(HomeboxInterface $homebox, AccountInterface $user) {
    if (!$homebox->getDashboardUserTabEnabled()) {
      return AccessResult::forbidden();
    }

    return $this->accessGeneral($homebox, $user);
  }

  /**
   * Provides the reference submission form.
   *
   * @param \Drupal\homebox\Entity\HomeboxInterface $homebox
   *   The homebox.
   *
   * @return array
   *   A homebox layout submission form.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function add(HomeboxInterface $homebox) {
    $entity = $this->entityTypeManager()->getStorage('homebox_layout')->create([
      'type' => $homebox->id(),
    ]);

    return $this->entityFormBuilder()->getForm($entity);
  }

  /**
   * The _title_callback for homebox routes.
   *
   * @param \Drupal\homebox\Entity\HomeboxInterface $homebox
   *   The homebox.
   *
   * @return string
   *   The page title.
   */
  public function pageTitle(HomeboxInterface $homebox) {
    return $homebox->label();
  }

}
