<?php

namespace Drupal\homebox\Form;

use Drupal\Component\Utility\Random;
use Drupal\Core\Access\CsrfTokenGenerator;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\CssCommand;
use Drupal\Core\Ajax\RemoveCommand;
use Drupal\Core\Block\BlockManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Layout\LayoutPluginManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\homebox\Entity\HomeboxInterface;
use Drupal\homebox\Entity\HomeboxLayout;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Serializer\Serializer;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Implements HomeboxDashboardForm for dashboard form.
 *
 * @ingroup homebox
 */
abstract class HomeboxDashboardForm extends FormBase {

  /**
   * Current user account class.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $account;

  /**
   * The current homebox entity.
   *
   * @var \Drupal\homebox\Entity\HomeboxInterface
   */
  protected $homebox;

  /**
   * Serializer service.
   *
   * @var \Symfony\Component\Serializer\Serializer
   */
  protected $serializer;

  /**
   * Request stack service.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * Plugin Manager Block service.
   *
   * @var \Drupal\Core\Block\BlockManagerInterface
   */
  protected $pluginManagerBlock;

  /**
   * Layout Plugin Manager Core service.
   *
   * @var \Drupal\Core\Layout\LayoutPluginManagerInterface
   */
  protected $pluginManagerCoreLayout;

  /**
   * The csrf service.
   *
   * @var \Drupal\Core\Access\CsrfTokenGenerator
   */
  protected $csrf;

  /**
   * HomeboxDashboardForm constructor.
   *
   * @param \Symfony\Component\Serializer\Serializer $serializer
   *   Serializer service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   Request Stack.
   * @param \Drupal\Core\Layout\LayoutPluginManagerInterface $plugin_manager_core_layout
   *   Core Layout of Plugin Manager.
   * @param \Drupal\Core\Block\BlockManagerInterface $plugin_manager_block
   *   Plugin manager block.
   * @param \Drupal\Core\Access\CsrfTokenGenerator $csrf
   *   CSRF Token generator.
   */
  public function __construct(
    Serializer $serializer,
    RequestStack $request_stack,
    LayoutPluginManagerInterface $plugin_manager_core_layout,
    BlockManagerInterface $plugin_manager_block,
    CsrfTokenGenerator $csrf
  ) {
    $this->serializer = $serializer;
    $this->requestStack = $request_stack;
    $this->pluginManagerCoreLayout = $plugin_manager_core_layout;
    $this->pluginManagerBlock = $plugin_manager_block;
    $this->csrf = $csrf;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    /** @var \Symfony\Component\Serializer\Serializer $serializer */
    $serializer = $container->get('serializer');
    /** @var \Symfony\Component\HttpFoundation\RequestStack $request_stack */
    $request_stack = $container->get('request_stack');
    /** @var \Drupal\Core\Layout\LayoutPluginManagerInterface $plugin_manager_core_layout */
    $plugin_manager_core_layout = $container->get('plugin.manager.core.layout');
    /** @var \Drupal\Core\Block\BlockManagerInterface $plugin_manager_block */
    $plugin_manager_block = $container->get('plugin.manager.block');
    /** @var \Drupal\Core\Access\CsrfTokenGenerator $csrf */
    $csrf = $container->get('csrf_token');
    return new static(
      $serializer,
      $request_stack,
      $plugin_manager_core_layout,
      $plugin_manager_block,
      $csrf
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'homebox_page';
  }

  /**
   * Save user settings.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   Current request.
   *
   * @return array
   *   The form structure.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public static function saveUserLayoutData(array $form, FormStateInterface &$form_state, Request $request) {
    $blocks = $form_state->getValue('blocks');

    if (!is_array($blocks)) {
      parse_str($form_state->getValue('blocks'), $blocks_value);
      $blocks = [];
      foreach ($blocks_value as $value) {
        parse_str($value, $block);
        // @todo Check block deleting in js. Possible there's parent element not remove.
        if ($block['id'] && $block['status']) {
          $blocks[] = $block;
        }
      }
    }

    // @todo We need to get the user here, the following implementation does not work for other users:
    // Get the user out of the request. If there is no user in the current
    // request, fallback to the current user
    // (needed for custom and default route).
    $user = $request->attributes->get('user') ?? \Drupal::currentUser();
    $homebox = $request->attributes->get('homebox');
    $homebox_layout_storage = \Drupal::entityTypeManager()->getStorage('homebox_layout');
    $homebox_layout_id = $homebox_layout_storage->getQuery()->accessCheck()
      ->condition('type', $homebox->id())
      ->condition('user_id', $user->id())
      ->execute();

    if (!empty($homebox_layout_id)) {
      /** @var \Drupal\Core\Entity\ContentEntityBase $homebox_layout */
      $homebox_layout = $homebox_layout_storage->load(array_shift($homebox_layout_id));
    }
    else {
      // Create homebox layout entity for current user.
      $values = [
        'user_id' => $user->id(),
        'name' => $homebox->id(),
        'status' => 1,
        'type' => $homebox->id(),
      ];
      /** @var \Drupal\Core\Entity\ContentEntityBase $homebox_layout */
      $homebox_layout = HomeboxLayout::create($values);
    }

    // Save user settings.
    $serializer = \Drupal::service('serializer');
    $data = $serializer->serialize($blocks, 'json');

    $homebox_layout->set('data', $data);
    $homebox_layout->set('layout_id', $homebox->getLayoutId());
    $homebox_layout->save();
    return $form;
  }

  /**
   * Ajax callback to hide block buttons.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   Response to hide buttons.
   */
  public static function hideBlockButtons() {
    $response = new AjaxResponse();
    $session = \Drupal::request()->getSession();
    $state = $session->get('block_buttons_visible');
    $response->addCommand(new CssCommand('.block-buttons', ['display' => $state ? 'none' : 'block']));
    $session->set('block_buttons_visible', !$state);
    return $response;
  }

  /**
   * Ajax callback to hide block.
   *
   * @param string $token
   *   Token of block.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   Response to hide buttons.
   */
  public function closeBlock($token) {
    $response = new AjaxResponse();
    $response->addCommand(new RemoveCommand('#' . $token));
    return $response;
  }

  /**
   * Ajax callback to add block.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   Current request.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   Response to hide buttons.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public static function addBlock(array $form, FormStateInterface $form_state, Request $request) {
    $user = \Drupal::currentUser();
    /** @var \Drupal\homebox\Entity\HomeboxInterface $homebox */
    $homebox = $request->attributes->get('homebox');

    // Get user layout settings or default homebox settings if not exists.
    /** @var \Drupal\homebox\Entity\HomeboxLayoutInterface $homebox_layout */
    $homebox_layout = self::getLayoutStorageData($homebox->id(), $user->id());
    $blocks = $homebox->getBlocks();
    if (!empty($homebox_layout)) {
      $serializer = \Drupal::service('serializer');
      $layout_blocks = $homebox_layout->get('data')->value;
      $layout_blocks = $serializer->decode($layout_blocks, 'json');
    }
    else {
      $layout_blocks = $blocks;
    }

    $element = $form_state->getTriggeringElement();

    foreach ($blocks as $block) {
      if ($block['id'] == $element['#attributes']['block_id']) {
        break;
      }
      $block = NULL;
    }

    if (!empty($block)) {
      $block['status'] = TRUE;
      $layout_blocks[] = $block;

      /** @var \Drupal\Core\Block\BlockPluginInterface $block_instance */
      $block_instance = \Drupal::service('plugin.manager.block')->createInstance($block['id']);
      $token = \Drupal::csrfToken()->get((new Random())->string(32));
      $render[$block['id']] = [
        '#type' => 'container',
        '#attributes' => [
          // @todo Where is this block_render class coming from?
          'class' => ['block-render', 'homebox-draggable'],
          'id' => $token,
        ],
      ];

      // @todo check if the following line is required (duplicate with later line).
      // Commented out until we're 100% sure to remove it:
      // $render[$block['id']]['render'] = $block_instance->build();
      $render[$block['id']]['render']['#theme'] = 'homebox_block';
      $render[$block['id']]['render']['#block_id'] = $block_instance->getPluginId();
      $render[$block['id']]['render']['#block_title'] = ($block['title'] != '' ? $block['title'] : $block_instance->label());
      $render[$block['id']]['render']['#block_content'] = $block_instance->build();

      $form['regions'][$block['region']][] = $render;
    }

    $form_state->setValue('blocks', $layout_blocks);

    self::saveUserLayoutData($form, $form_state, $request);

    return $form['regions'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, HomeboxInterface $homebox = NULL, AccountInterface $account = NULL) {
    $this->homebox = $homebox;
    if ($account === NULL) {
      $account = $this->currentUser();
    }
    $this->account = $account;

    $is_ajax = $this->getRequest()->isXmlHttpRequest();
    if (!$is_ajax) {
      $session = $this->requestStack->getCurrentRequest()->getSession();
      $session->set('block_buttons_visible', FALSE);
    }

    // Get user layout settings or default homebox settings if not exists.
    /** @var \Drupal\homebox\Entity\HomeboxLayoutInterface $homebox_layout */
    $homebox_layout = $this->getLayoutStorageData($homebox->id(), $account->id());
    $blocks_available_for_adding = $homebox->getBlocks();
    // Check if user layout settings exists and haven't been changed.
    $is_layout_id = !empty($homebox_layout) && isset($homebox_layout->get('layout_id')->value);
    if (!empty($homebox_layout) && $is_layout_id && $homebox->getLayoutId() == $homebox_layout->get('layout_id')->value) {
      $blocks = $homebox_layout->get('data')->value;
      $blocks = $this->serializer->decode($blocks, 'json');
    }
    else {
      $blocks = $blocks_available_for_adding;
    }

    if (empty($homebox->getBlocks())) {
      $form['no_blocks_set'] = [
        '#type' => 'markup',
        '#markup' => $this->t('There are no blocks set! If you have the right permissions you can visit the current Homebox <a href="@link">Settings Form </a> to set the Blocks', ['@link' => Url::fromRoute('homebox.settings_form', ['homebox' => $homebox->id()])->toString()]),
      ];
      return $form;
    }

    // Get layout regions.
    /** @var \Drupal\Core\Layout\LayoutInterface $layout_instance */
    $layout_instance = $this->pluginManagerCoreLayout->createInstance($homebox->getLayoutId(), []);

    $render = $regions = [];

    $form['#attached']['library'][] = 'homebox/draggable-blocks';
    $form['blocks'] = [
      '#type' => 'hidden',
    ];
    $form['add_block_button'] = [
      '#type' => 'button',
      '#value' => $this->t('Add a block'),
      '#ajax' => [
        'callback' => '::hideBlockButtons',
        'event' => 'click',
      ],
      '#attributes' => [
        'class' => [
          'button--add',
        ],
      ],
    ];
    $form['reset_block_button'] = [
      '#type' => 'link',
      '#title' => $this->t('Reset to defaults'),
      '#url' => Url::fromRoute(
        'homebox.dashboard_reset',
        [
          'homebox' => $this->homebox->id(),
          'user' => $this->account->id(),
        ],
        [
          'query' => $this->getDestinationArray(),
        ],
      ),
      '#attributes' => [
        'class' => [
          'button',
          'button--primary',
          'button--reset',
        ],
      ],

    ];
    $form['save_form'] = [
      '#type' => 'button',
      '#value' => $this->t('Save'),
      '#attributes' => [
        'class' => ['homebox-save-form'],
      ],
      '#ajax' => [
        'callback' => '::saveUserLayoutData',
        'event' => 'click',
      ],
    ];
    $form['block_buttons'] = [
      '#type' => 'fieldset',
      '#attributes' => [
        'class' => ['block-buttons'],
        'style' => 'display: none;',
      ],
    ];

    // Create constant list of blocks available for adding on homebox page.
    foreach ($blocks_available_for_adding as $block) {
      /** @var \Drupal\Core\Block\BlockPluginInterface $block_instance */
      $block_instance = $this->pluginManagerBlock->createInstance($block['id']);

      $form['block_buttons']['add_block_' . $block['id']] = [
        '#type' => 'button',
        '#value' => (isset($block['title']) && $block['title'] != '' ? $block['title'] : $block_instance->label()),
        '#attributes' => [
          'block_id' => $block['id'],
        ],
        '#ajax' => [
          'callback' => '::addBlock',
          'wrapper' => 'homebox',
          'event' => 'click',
        ],
      ];
    }

    foreach ($blocks as $block) {
      $block_instance = $this->pluginManagerBlock->createInstance($block['id']);
      if ($block['status']) {
        $token = $this->csrf->get((new Random())->string(32));
        $render[$block['id']] = [
          '#type' => 'container',
          '#attributes' => [
            'id' => $token,
            'class' => ['homebox-draggable'],
          ],
        ];

        // @todo check if the following line is required (duplicate with later line).
        // Commented out until we're 100% sure to remove it:
        // $render[$block['id']]['render'] = $block_instance->build();
        $render[$block['id']]['render']['#theme'] = 'homebox_block';
        $render[$block['id']]['render']['#block_id'] = $block_instance->getPluginId();
        $render[$block['id']]['render']['#block_title'] = (isset($block['title']) && $block['title'] != '' ? $block['title'] : $block_instance->label());
        $render[$block['id']]['render']['#block_content'] = $block_instance->build();

        $regions[$block['region']][] = $render[$block['id']];
        $regions[$block['region']]['#attributes'] = [
          'id' => 'region_' . $block['region'],
          'class' => ['homebox-column', 'js-homebox-region'],
        ];
      }
    }

    $plugin_definition = $layout_instance->getPluginDefinition();
    foreach ($plugin_definition->getRegions() as $id => $region) {
      if (!isset($regions[$id])) {
        $regions[$id] = [
          '#markup' => '',
          '#attributes' => [
            'id' => 'region_' . $id,
            'class' => ['homebox-column', 'js-homebox-region'],
          ],
        ];
      }
    }

    $form['regions'] = $layout_instance->build($regions);
    // Set ID:
    $form['regions']['#attributes']['id'] = 'homebox';

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
  }

  /**
   * Helper function to get the layout storage data.
   *
   * @param int $homebox
   *   Current homebox id.
   * @param int $user_id
   *   Current user id.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   Layout homebox of current user.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public static function getLayoutStorageData($homebox, $user_id) {
    $homebox_layout_storage = \Drupal::entityTypeManager()->getStorage('homebox_layout');
    $homebox_layout_id = $homebox_layout_storage->getQuery()->accessCheck()
      ->condition('type', $homebox)
      ->condition('user_id', $user_id)
      ->execute();

    if (!empty($homebox_layout_id)) {
      $homebox_layout = $homebox_layout_storage->load(array_shift($homebox_layout_id));
    }
    else {
      $homebox_layout = NULL;
    }

    return $homebox_layout;
  }

}
