<?php

namespace Drupal\homebox\Form;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Form\ConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\homebox\Entity\HomeboxInterface;
use Drupal\homebox\Entity\HomeboxLayout;

/**
 * Provides a confirmation form before clearing out the examples.
 */
class HomeboxDashboardResetForm extends ConfirmFormBase {

  /**
   * The homebox entity to reset.
   *
   * @var \Drupal\homebox\Entity\HomeboxInterface
   */
  protected $homebox;

  /**
   * The user which owns the homebox.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $account;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'homebox_homebox_dashboard_reset';
  }

  /**
   * {@inheritdoc}
   */
  public function getQuestion() {
    return $this->t('Are you sure you want to do this?');
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelUrl() {
    return Url::fromRoute('homebox.dashboard_page', [
      'homebox' => $this->homebox->id(),
      'user' => $this->account->id(),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, HomeboxInterface $homebox = NULL, AccountInterface $account = NULL) {
    if ($account === NULL) {
      $account = $this->currentUser();
    }
    $this->homebox = $homebox;
    $this->account = $account;
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $homebox = $this->homebox;
    $account = $this->account;
    $originalBlocks = $homebox->getBlocks();
    $homebox_layout_storage = \Drupal::entityTypeManager()->getStorage('homebox_layout');
    $homebox_layout_id = $homebox_layout_storage->getQuery()->accessCheck()
      ->condition('type', $homebox->id())
      ->condition('user_id', $account->id())
      ->execute();

    if (!empty($homebox_layout_id)) {
      /** @var \Drupal\Core\Entity\ContentEntityBase $homebox_layout */
      $homebox_layout = $homebox_layout_storage->load(array_shift($homebox_layout_id));
    }
    else {
      // Create homebox layout entity for current user.
      $values = [
        'user_id' => $account->id(),
        'name' => $homebox->id(),
        'status' => 1,
        'type' => $homebox->id(),
      ];
      /** @var \Drupal\Core\Entity\ContentEntityBase $homebox_layout */
      $homebox_layout = HomeboxLayout::create($values);
    }

    // Save user settings.
    $serializer = \Drupal::service('serializer');
    $data = $serializer->serialize($originalBlocks, 'json');

    $homebox_layout->set('data', $data);
    $homebox_layout->set('layout_id', $homebox->getLayoutId());
    // @todo Ensure this creates a new revision (with reset description), if revisioning is enabled.
    $homebox_layout->save();

    if ($this->currentUser()->id() != $this->account->id()) {
      $this->messenger()->addStatus($this->t('The homebox @homebox of user @user was reset!', [
        '@homebox' => $this->homebox->id(),
        '@user' => $this->account->getAccountName(),
      ]));
    }
    else {
      $this->messenger()->addStatus($this->t('Your homebox @homebox was resettet!', ['@homebox' => $this->homebox->id()]));
    }
    $query = $this->getRequest()->query;
    // Success url = cancel url:
    $url = $this->getCancelUrl();
    if ($query->has('destination')) {
      $options = UrlHelper::parse($query->get('destination'));
      // @todo Revisit this in https://www.drupal.org/node/2418219.
      try {
        $url = Url::fromUserInput('/' . ltrim($options['path'], '/'), $options);
      }
      catch (\InvalidArgumentException $e) {
        throw $e;
      }
    }
    $form_state->setRedirectUrl($url);
  }

}
