<?php

namespace Drupal\homebox\Form;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Path\PathValidator;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Layout\LayoutPluginManager;
use Drupal\Core\Block\BlockManagerInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\ProxyClass\Routing\RouteBuilder;
use Drupal\Core\Plugin\CachedDiscoveryClearerInterface;
use Drupal\Core\Plugin\Context\ContextRepositoryInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;

/**
 * General settings for a Homebox entity.
 */
class HomeboxForm extends EntityForm {

  /**
   * The condition plugin manager.
   *
   * @var \Drupal\Core\Condition\ConditionManager
   */
  protected $manager;

  /**
   * The layout plugin manager.
   *
   * @var \Drupal\Core\Layout\LayoutPluginManager
   */
  protected $layoutPluginManager;

  /**
   * Manager block plugin.
   *
   * @var \Drupal\Core\Block\BlockManagerInterface
   */
  protected $blockManager;

  /**
   * Manager block plugin.
   *
   * @var \Drupal\Core\Plugin\Context\ContextRepositoryInterface
   */
  protected $contextRepository;

  /**
   * A plugin cache clear instance.
   *
   * @var \Drupal\Core\Plugin\CachedDiscoveryClearerInterface
   */
  protected $pluginCacheClearer;

  /**
   * A cache backend interface instance.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cacheRender;

  /**
   * An event dispatcher.
   *
   * @var \Symfony\Contracts\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The route builder instance.
   *
   * @var \Drupal\Core\ProxyClass\Routing\RouteBuilder
   */
  protected $routeBuilder;

  /**
   * The path validator instance.
   *
   * @var \Drupal\Core\Path\PathValidator
   */
  protected $pathValidator;

  /**
   * HomeboxForm constructor.
   *
   * @param \Drupal\Core\Layout\LayoutPluginManager $layoutPluginManager
   *   Layout plugin manager.
   * @param \Drupal\Core\Block\BlockManagerInterface $block_manager
   *   Manager block plugin.
   * @param \Drupal\Core\Plugin\Context\ContextRepositoryInterface $context_repository
   *   Context repository plugin.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cacheRender
   *   A cache backend interface instance.
   * @param \Drupal\Core\Plugin\CachedDiscoveryClearerInterface $plugin_cache_clearer
   *   A plugin cache clear instance.
   * @param \Symfony\Contracts\EventDispatcher\EventDispatcherInterface $eventDispatcher
   *   A path alias manager instance.
   * @param \Drupal\Core\ProxyClass\Routing\RouteBuilder $routeBuilder
   *   The route builder instance.
   * @param \Drupal\Core\Path\PathValidator $pathValidator
   *   The path validator instance.
   */
  public function __construct(LayoutPluginManager $layoutPluginManager, BlockManagerInterface $block_manager, ContextRepositoryInterface $context_repository, CacheBackendInterface $cacheRender, CachedDiscoveryClearerInterface $plugin_cache_clearer, EventDispatcherInterface $eventDispatcher, RouteBuilder $routeBuilder, PathValidator $pathValidator) {
    $this->layoutPluginManager = $layoutPluginManager;
    $this->blockManager = $block_manager;
    $this->contextRepository = $context_repository;
    $this->cacheRender = $cacheRender;
    $this->pluginCacheClearer = $plugin_cache_clearer;
    $this->eventDispatcher = $eventDispatcher;
    $this->routeBuilder = $routeBuilder;
    $this->pathValidator = $pathValidator;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    /** @var \Drupal\Core\Plugin\CachedDiscoveryClearerInterface $plugin_cache_clearer */
    $plugin_cache_clearer = $container->get('plugin.cache_clearer');
    /** @var \Drupal\Core\Cache\CacheBackendInterface $cache_render */
    $cache_render = $container->get('cache.render');
    /** @var \Drupal\Core\Plugin\Context\ContextRepositoryInterface $context_repository */
    $context_repository = $container->get('context.repository');
    /** @var \Drupal\Core\Block\BlockManagerInterface $block_manager */
    $block_manager = $container->get('plugin.manager.block');
    /** @var \Drupal\Core\Layout\LayoutPluginManager $layout_plugin_manager */
    $layout_plugin_manager = $container->get('plugin.manager.core.layout');
    /** @var \Symfony\Contracts\EventDispatcher\EventDispatcherInterface $eventDispatcher*/
    $eventDispatcher = $container->get('event_dispatcher');
    /** @var \Drupal\Core\ProxyClass\Routing\RouteBuilder $route_builder*/
    $route_builder = $container->get('router.builder');
    /** @var \Drupal\Core\Path\PathValidator $path_validator */
    $path_validator = $container->get('path.validator');
    return new static(
      $layout_plugin_manager,
      $block_manager,
      $context_repository,
      $cache_render,
      $plugin_cache_clearer,
      $eventDispatcher,
      $route_builder,
      $path_validator
    );
  }

  /**
   * Provide layout plugin instance.
   *
   * @param string $layout_id
   *   Layout id.
   * @param array $layout_settings
   *   Layout settings.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state object.
   *
   * @return mixed|object
   *   Layout plugin.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  protected function getLayout($layout_id, array $layout_settings, FormStateInterface $form_state) {
    if (!$layout_plugin = $form_state->get('layout_plugin')) {
      $layout_plugin = $this->layoutPluginManager->createInstance($layout_id, $layout_settings);
      $form_state->set('layout_plugin', $layout_plugin);
    }

    return $layout_plugin;
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);

    /** @var \Drupal\homebox\Entity\HomeboxInterface $homebox */
    $homebox = $this->entity;

    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Homebox Label'),
      '#maxlength' => 255,
      '#default_value' => $homebox->label(),
      '#description' => $this->t("Label for the current homebox entity."),
      '#required' => TRUE,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $homebox->getOriginalId(),
      '#machine_name' => [
        'exists' => '\Drupal\homebox\Entity\Homebox::load',
      ],
      '#disabled' => !$homebox->isNew(),
    ];

    $form['homebox_vertical_tabs'] = [
      '#type' => 'vertical_tabs',
      '#title' => $this->t('Settings'),
    ];

    $form['general_form'] = [
      '#type' => 'details',
      '#title' => $this->t('General'),
      '#description' => $this->t('General homebox settings'),
      '#group' => 'homebox_vertical_tabs',
    ];

    $form['general_form']['layoutId'] = [
      '#type' => 'select',
      '#title' => $this->t('Select Layout'),
      '#options' => $this->layoutPluginManager->getLayoutOptions(),
      '#description' => $this->t('Set the column layout you want for this Homebox page. Note, that changing this will RESET all set blocks!'),
      '#default_value' => $homebox->getLayoutId(),
      '#required' => TRUE,
    ];

    $form['general_form']['restrictedRoles'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Restrict access to this homebox to certain roles'),
      '#options' => $this->getRolesList(),
      '#default_value' => $homebox->getRestrictedRoles(),
      '#description' => $this->t('If at least one role is selected, only these roles may access this homebox. For this to take effect, you also have to enable the global "Use Homebox" permission for these roles. Users with role "Bypass Homebox access control" bypass this setting. If no roles are selected, all users with "Use homebox" permission have access.'),
    ];

    $form['page_form'] = [
      '#type' => 'details',
      '#title' => $this->t('Dashboard Page'),
      '#group' => 'homebox_vertical_tabs',
    ];

    $form['page_form']['dashboardPageEnabled'] = [
      '#type' => 'checkbox',
      '#title' => 'Enable',
      '#default_value' => $homebox->getDashboardPageEnabled(),
      '#description' => $this->t("Enables a homebox dashboard page. You can access the page under homebox/{homebox_id}/{user-id}. If you don't specify a 'user-id' it defaults to your own user-id."),
    ];

    $form['page_form']['customPagePath'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Custom dashboard path'),
      '#maxlength' => 255,
      '#default_value' => $homebox->getCustomPagePath(),
      '#description' => $this->t("Specify a URL by which this page can be accessed from. Use a relative path, for example '/dashboard'."),
      '#element_validate' => [[$this, 'validatePath']],
      '#states' => [
        'visible' => [
          ':input[name="dashboardPageEnabled"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $form['user_tab_form'] = [
      '#type' => 'details',
      '#title' => $this->t('Dashboard User Tab'),
      '#group' => 'homebox_vertical_tabs',
    ];

    $form['user_tab_form']['dashboardUserTabEnabled'] = [
      '#type' => 'checkbox',
      '#title' => 'Enable',
      '#default_value' => $homebox->getDashboardUserTabEnabled(),
      '#description' => $this->t('Enable a homebox dashboard user tab page.'),
    ];

    // @todo Implement this feature for 3.x (https://www.drupal.org/project/homebox/issues/3266889)
    // $form['block_form'] = [
    // '#type' => 'details',
    // '#title' => $this->t('Dashboard Block'),
    // '#group' => 'homebox_vertical_tabs',
    // ];
    // $form['block_form']['block_form_text'] = [
    // '#type' => 'markup',
    // '#markup' => 'Future implementation:
    // Homebox also alternatively provides a block.
    // Simply drag it into the reserved region to enable it.',
    // ];
    return $form;
  }

  /**
   * Helper function to validate the 'path' form element.
   *
   * @param array $element
   *   The form element to validate.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function validatePath(array &$element, FormStateInterface $form_state) {
    if (!empty($form_state->getValue('dashboardPageEnabled')) && !empty($element['#value'])) {
      /** @var \Drupal\homebox\Entity\HomeboxInterface $homebox */
      $homebox = $this->entity;
      // Ensure the path has a leading slash.
      if ($element['#value'][0] !== '/') {
        $form_state->setErrorByName('customPagePath', $this->t('Manually entered paths should start with a leading slash ("/").'));
      }
      if ($value = trim($element['#value'], '/')) {
        $value = '/' . $value;
        $form_state->setValueForElement($element, $value);

        // Ensure that the path is unique:
        if ($this->pathValidator->getUrlIfValidWithoutAccessCheck($value) !== FALSE && $homebox->getCustomPagePath() != $value) {
          $form_state->setErrorByName('customPagePath', $this->t('The path exists already, and is therefore not unique.'));
        }
      }
      // Check to make sure the path exists after stripping slashes.
      else {
        $form_state->setErrorByName('customPagePath', $this->t("Path is required."));
      }
    }
  }

  /**
   * Retreives the user roles.
   *
   * @return array
   *   The user array.
   */
  public function getRolesList(): array {
    $roles = user_roles();
    $role_list = [];
    foreach ($roles as $role) {
      if (!$role->isAdmin() && $role->id() !== 'anonymous') {
        $role_list[$role->id()] = $role->label();
      }
    }
    return $role_list;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\homebox\Entity\HomeboxInterface $homebox */
    $homebox = $this->entity;
    // Get previous layoutId:
    $oldLayoutId = $form['general_form']['layoutId']['#default_value'] ?? NULL;
    // Empty blocks if layout has changed:
    if ($form_state->getValue('layoutId') != $oldLayoutId) {
      $this->messenger()->addMessage($this->t('Removed all blocks, as @label Homebox layout was changed.', [
        '@label' => $homebox->label(),
      ]), MessengerInterface::TYPE_STATUS, TRUE);
      $homebox->setBlocks([]);
    }
    $status = parent::save($form, $form_state);
    switch ($status) {
      case SAVED_NEW:
        $this->messenger()->addMessage($this->t('Created the @label Homebox.', [
          '@label' => $homebox->label(),
        ]));
        break;

      default:
        $this->messenger()->addMessage($this->t('Saved the @label Homebox.', [
          '@label' => $homebox->label(),
        ]));
    }
    // We need to let our RouteSubscriber know,
    // that there could be a custom path change:
    $this->routeBuilder->setRebuildNeeded();
    // Clear cache to rebuild tasks links.
    $this->pluginCacheClearer->clearCachedDefinitions();
    $this->cacheRender->invalidateAll();
    $form_state->setRedirectUrl($homebox->toUrl('collection'));
    return $status;
  }

}
